#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <curses.h>
#include <term.h>

#include <readline/readline.h>
#include <readline/history.h>

#include "execpty.h"

/* screenwidth is a global variable in the readline library that we're
   not supposed to know about, but comes in very useful here. */
extern int screenwidth;

/* The fd of the pty master. */
static int ule_ptyfd;

/* Our "prompt", which consists of anything the child process has
   written after the last newline. */
static char *promptbuf = NULL;
static size_t promptstart = 0;
static size_t promptlen = 0;
static size_t promptbuflen = 0;
static size_t promptbufsize = 0;

static struct {
    const char *highlight;
    const char *nohighlight;
    const char *cr;
    const char *ce;
} tcap;

static void
nomorehighlight(void)
{
    if (tcap.nohighlight) {
	tputs(tcap.nohighlight, 1, putchar);
	fflush(stdout);
    }
}

static void
init_termcapstuff(char hl)
{
    static char term_buf[4096];
    static char tcap_buf[2048];

    const char *term;
    char *b = tcap_buf;
    
    term = getenv("TERM");
    if (term == NULL) {
	fprintf(stderr, "ule: TERM not set\n");
	exit(1);
    }
    if (tgetent(term_buf, term) != 1) {
	fprintf(stderr, "ule: unknown TERM %s\n", term);
	exit(1);
    }

    switch (hl) {
      case 's':
	tcap.highlight = tgetstr("so", &b);
	tcap.nohighlight = tgetstr("se", &b);
	break;
      case 'f':
	tcap.highlight = tgetstr("mb", &b);
	tcap.nohighlight = tgetstr("me", &b);
	break;
      case 'b':
	tcap.highlight = tgetstr("md", &b);
	tcap.nohighlight = tgetstr("me", &b);
	break;
      case 'r':
	tcap.highlight = tgetstr("mr", &b);
	tcap.nohighlight = tgetstr("me", &b);
	break;
      case 'u':
	tcap.highlight = tgetstr("us", &b);
	tcap.nohighlight = tgetstr("ue", &b);
      default:
	tcap.highlight = NULL;
	tcap.nohighlight = NULL;
    }

    tcap.cr = tgetstr("cr", &b);
    tcap.ce = tgetstr("ce", &b);
}

static int
addtopromptbuf(int c)
{
    if (promptbuflen >= promptbufsize) {
	promptbufsize = promptbuflen + 1;
	promptbuf = realloc(promptbuf, promptbufsize);
	if (promptbuf == NULL) {
	    fprintf(stderr, "ule: out of memory\n");
	    exit(1);
	}
    }
    promptbuf[promptbuflen++] = c;
    return c;
}

static void
terminate_promptbuf(void)
{
    promptbuflen = promptstart + promptlen;
    if (tcap.highlight) {
	addtopromptbuf(RL_PROMPT_START_IGNORE);
	tputs(tcap.highlight, 1, addtopromptbuf);
	addtopromptbuf(RL_PROMPT_END_IGNORE);
    }
    addtopromptbuf('\0');
}

static void
initiate_promptbuf(void)
{
    promptbuflen = 0;
    if (tcap.nohighlight) {
	addtopromptbuf(RL_PROMPT_START_IGNORE);
	tputs(tcap.nohighlight, 1, addtopromptbuf);
	addtopromptbuf(RL_PROMPT_END_IGNORE);
    }
    promptstart = promptbuflen;
    promptlen = 0;
    terminate_promptbuf();
}

static void
do_inputline(char *line)
{
    size_t linelen;
    char ret = '\r';
    struct iovec iov[2];
    
    if (line == NULL) {
	exit(0);		/* EOF on stdin */
    }

    linelen = strlen(line);
    iov[0].iov_base = line;
    iov[0].iov_len = linelen;
    iov[1].iov_base = &ret;
    iov[1].iov_len = 1;

    setslavenoecho();
    if (writev(ule_ptyfd, iov, 2) < 0) {
	perror("ule: writev to child");
	exit(1);
    }

    if (linelen > 0) {
	add_history(line);
    }
    free(line);

    rl_callback_handler_remove();
    initiate_promptbuf();
    rl_callback_handler_install(promptbuf, do_inputline);
}

static void
finalize_readline(void)
{
    rl_callback_handler_remove();
    nomorehighlight();
}

void
initialize_readline(int ptyfd, char hl)
{
    ule_ptyfd = ptyfd;
 
    init_termcapstuff(hl);
    initiate_promptbuf();
    rl_callback_handler_install(promptbuf, do_inputline);
    atexit(finalize_readline);
}

void
do_stdinput(void)
{
    rl_callback_read_char();
}

static void
clear_current_line(void)
{
    if (tcap.cr && tcap.ce) {
	tputs(tcap.cr, 1, putchar);
	tputs(tcap.ce, 1, putchar);
	fflush(stdout);
    } else {
	int width;
	char linebuf[1024];
	
	width = screenwidth;
	if (width >= sizeof linebuf) {
	    width = sizeof linebuf - 1;
	}
	memset(linebuf, ' ', width + 1);
	linebuf[0] = '\r';
	linebuf[width + 1] = '\r';
	write(STDOUT_FILENO, linebuf, width + 2);
    }
}

void
do_childinput(void)
{
    char buf[65536];		/* Arbitrary size */
    ssize_t numread;
    int point;
    HISTORY_STATE *history_state;
    int histcount;
    char *savearea;
    const char *ret;
    size_t retoffset;
    size_t tocopy;

    numread = read(ule_ptyfd, buf, sizeof buf);
    if (numread <= 0) {
	exit(0);
    }

    /* Save current editing state. */
    point = rl_point;
    history_state = history_get_history_state();
    histcount = history_state->length - history_state->offset;
    free(history_state);
    rl_get_next_history(histcount, 0);
    savearea = strdup(rl_line_buffer);
    if (savearea == NULL) {
	fprintf(stderr, "ule: out of memory\n");
	exit(1);
    }

    /* Get rid of current input line. */
    rl_delete_text(0, rl_end);
    rl_redisplay();
    rl_callback_handler_remove();
    nomorehighlight();

    /* Create new prompt. */
    ret = memchr(buf, '\r', numread);
    if (ret == NULL) {
	retoffset = 0;
	clear_current_line();
    } else {
	ssize_t numlook;
	const char *newret;
	retoffset = ret - buf;
	numlook = numread - retoffset;
	while (numlook > 1 && (newret = memchr(buf + retoffset + 1, '\r', numlook - 1))) {
	    size_t newretoff = newret - buf;
	    numlook -= newretoff - retoffset;
	    retoffset = newretoff;
	}
	initiate_promptbuf();
    }
    while (retoffset < numread && 
	   (buf[retoffset] == '\r' || buf[retoffset] == '\n')) ++retoffset;
    write(STDOUT_FILENO, buf, retoffset);

    tocopy = numread - retoffset;
    if (promptbuflen + tocopy > promptbufsize) {
	promptbufsize += tocopy;
	promptbuf = realloc(promptbuf, promptbufsize);
	if (promptbuf == NULL) {
	    fprintf(stderr, "ule: out of memory\n");
	    exit(1);
	}
    }
    memcpy(promptbuf + promptstart + promptlen, buf + retoffset, tocopy);
    promptlen += tocopy;

    /* Take complete lines off the front end of the prompt. */
    while (promptlen >= screenwidth) {
	write(STDOUT_FILENO, promptbuf + promptstart, screenwidth);
	promptlen -= screenwidth;
	promptbuflen -= screenwidth;
	memmove(promptbuf + promptstart, promptbuf + promptstart + screenwidth, promptbuflen - promptstart);
    }
    terminate_promptbuf();

    /* Reinstall callback with new prompt. */
    rl_callback_handler_install(promptbuf, do_inputline);

    /* Restore editing state. */
    rl_insert_text(savearea);
    free(savearea);
    rl_get_previous_history(histcount, 0);
    rl_point = point;
    rl_redisplay();
}
